@@ -135,6 +135,12 @@ def check_config():
     return encode_data(get_storage().check_config(**decode_request(request)))
 
 
+@app.route('/reset', methods=['POST'])
+@timed
+def reset():
+    return encode_data(get_storage().reset(**decode_request(request)))
+
+
 @app.route('/content/missing', methods=['POST'])
 @timed
 def content_missing():
@@ -149,6 +155,13 @@ def content_missing_per_sha1():
         **decode_request(request)))
 
 
+@app.route('/content/skipped/missing', methods=['POST'])
+@timed
+def skipped_content_missing():
+    return encode_data(get_storage().skipped_content_missing(
+        **decode_request(request)))
+
+
 @app.route('/content/present', methods=['POST'])
 @timed
 def content_find():
@@ -293,7 +306,8 @@ def object_find_by_sha1_git():
 @encode
 @process_metrics
 def snapshot_add():
-    return get_storage().snapshot_add(**decode_request(request))
+    req_data = decode_request(request)
+    return get_storage().snapshot_add(**req_data)
 
 
 @app.route('/snapshot', methods=['POST'])
@@ -380,6 +394,13 @@ def origin_visit_get():
         **decode_request(request)))
 
 
+@app.route('/origin/visit/find_by_date', methods=['POST'])
+@timed
+def origin_visit_find_by_date():
+    return encode_data(get_storage().origin_visit_find_by_date(
+        **decode_request(request)))
+
+
 @app.route('/origin/visit/getby', methods=['POST'])
 @timed
 def origin_visit_get_by():
@@ -387,6 +408,13 @@ def origin_visit_get_by():
         get_storage().origin_visit_get_by(**decode_request(request)))
 
 
+@app.route('/origin/visit/get_latest', methods=['POST'])
+@timed
+def origin_visit_get_latest():
+    return encode_data(
+        get_storage().origin_visit_get_latest(**decode_request(request)))
+
+
 @app.route('/origin/visit/add', methods=['POST'])
 @timed
 @encode
@@ -404,6 +432,13 @@ def origin_visit_update():
         **decode_request(request)))
 
 
+@app.route('/origin/visit/upsert', methods=['POST'])
+@timed
+def origin_visit_upsert():
+    return encode_data(get_storage().origin_visit_upsert(
+        **decode_request(request)))
+
+
 @app.route('/person', methods=['POST'])
 @timed
 def person_get():
