@@ -14,6 +14,7 @@ class Db(BaseDb):
     """Proxy to the SWH DB, with wrappers around stored procedures
 
     """
+
     def mktemp_dir_entry(self, entry_type, cur=None):
         self._cursor(cur).execute('SELECT swh_mktemp_dir_entry(%s)',
                                   (('directory_entry_%s' % entry_type),))
@@ -124,11 +125,22 @@ class Db(BaseDb):
             SELECT 1 FROM content c WHERE c.sha1 = t.sha1
         )""", ((sha1,) for sha1 in sha1s))
 
-    def skipped_content_missing_from_temp(self, cur=None):
+    def skipped_content_missing(self, contents, cur=None):
+        if not contents:
+            return []
         cur = self._cursor(cur)
 
-        cur.execute("""SELECT sha1, sha1_git, sha256, blake2s256
-                       FROM swh_skipped_content_missing()""")
+        query = """SELECT * FROM (VALUES %s) AS t (%s)
+                   WHERE not exists
+                   (SELECT 1 FROM skipped_content s WHERE
+                       s.sha1 is not distinct from t.sha1 and
+                       s.sha1_git is not distinct from t.sha1_git and
+                       s.sha256 is not distinct from t.sha256);""" % \
+                ((', '.join('%s' for _ in contents)),
+                 ', '.join(self.content_hash_keys))
+        cur.execute(query,
+                    [tuple(cont[key] for key in self.content_hash_keys)
+                     for cont in contents])
 
         yield from cur
 
@@ -206,16 +218,23 @@ class Db(BaseDb):
         """
         cur = self._cursor(cur)
 
+        checksum_dict = {'sha1': sha1, 'sha1_git': sha1_git,
+                         'sha256': sha256, 'blake2s256': blake2s256}
+        where_parts = []
+        args = []
+        # Adds only those keys which have value other than None
+        for algorithm in checksum_dict:
+            if checksum_dict[algorithm] is not None:
+                args.append(checksum_dict[algorithm])
+                where_parts.append(algorithm + '= %s')
+        query = ' AND '.join(where_parts)
         cur.execute("""SELECT %s
-                       FROM swh_content_find(%%s, %%s, %%s, %%s)
-                       LIMIT 1""" % ','.join(self.content_find_cols),
-                    (sha1, sha1_git, sha256, blake2s256))
-
-        content = cur.fetchone()
-        if set(content) == {None}:
-            return None
-        else:
-            return content
+                       FROM content WHERE %s
+                       """
+                    % (','.join(self.content_find_cols),  query),
+                    args)
+        content = cur.fetchall()
+        return content
 
     def directory_missing_from_list(self, directories, cur=None):
         cur = self._cursor(cur)
@@ -282,21 +301,22 @@ class Db(BaseDb):
     revision_get_cols = revision_add_cols + [
         'author_id', 'committer_id', 'parents']
 
-    def origin_visit_add(self, origin, ts, cur=None):
+    def origin_visit_add(self, origin, ts, type, cur=None):
         """Add a new origin_visit for origin origin at timestamp ts with
         status 'ongoing'.
 
         Args:
             origin: origin concerned by the visit
             ts: the date of the visit
+            type: type of loader for the visit
 
         Returns:
             The new visit index step for that origin
 
         """
         cur = self._cursor(cur)
-        self._cursor(cur).execute('SELECT swh_origin_visit_add(%s, %s)',
-                                  (origin, ts))
+        self._cursor(cur).execute('SELECT swh_origin_visit_add(%s, %s, %s)',
+                                  (origin, ts, type))
         return cur.fetchone()[0]
 
     def origin_visit_update(self, origin_id, visit_id, updates, cur=None):
@@ -327,8 +347,21 @@ class Db(BaseDb):
         })
         cur.execute(query, (*values, *where_values))
 
-    origin_visit_get_cols = ['origin', 'visit', 'date', 'status', 'metadata',
-                             'snapshot']
+    def origin_visit_upsert(self, origin, visit, date, type, status,
+                            metadata, snapshot, cur=None):
+        cur = self._cursor(cur)
+        query = """INSERT INTO origin_visit ({cols}) VALUES ({values})
+                   ON CONFLICT ON CONSTRAINT origin_visit_pkey DO
+                   UPDATE SET {updates}""".format(
+                cols=', '.join(self.origin_visit_get_cols),
+                values=', '.join('%s' for col in self.origin_visit_get_cols),
+                updates=', '.join('{0}=excluded.{0}'.format(col)
+                                  for col in self.origin_visit_get_cols))
+        cur.execute(
+            query, (origin, visit, date, type, status, metadata, snapshot))
+
+    origin_visit_get_cols = ['origin', 'visit', 'date', 'type', 'status',
+                             'metadata', 'snapshot']
 
     def origin_visit_get_all(self, origin_id,
                              last_visit=None, limit=None, cur=None):
@@ -388,6 +421,15 @@ class Db(BaseDb):
             return None
         return r[0]
 
+    def origin_visit_find_by_date(self, origin, visit_date, cur=None):
+        cur = self._cursor(cur)
+        cur.execute(
+            'SELECT * FROM swh_visit_find_by_date(%s, %s)',
+            (origin, visit_date))
+        r = cur.fetchall()
+        if r:
+            return r[0]
+
     def origin_visit_exists(self, origin_id, visit_id, cur=None):
         """Check whether an origin visit with the given ids exists"""
         cur = self._cursor(cur)
@@ -398,34 +440,40 @@ class Db(BaseDb):
 
         return bool(cur.fetchone())
 
-    def origin_visit_get_latest_snapshot(self, origin_id,
-                                         allowed_statuses=None,
-                                         cur=None):
-        """Retrieve the most recent origin_visit which references a snapshot
+    def origin_visit_get_latest(
+            self, origin_id, allowed_statuses=None, require_snapshot=False,
+            cur=None):
+        """Retrieve the most recent origin_visit of the given origin,
+        with optional filters.
 
         Args:
             origin_id: the origin concerned
             allowed_statuses: the visit statuses allowed for the returned visit
+            require_snapshot (bool): If True, only a visit with a known
+                snapshot will be returned.
 
         Returns:
             The origin_visit information, or None if no visit matches.
         """
         cur = self._cursor(cur)
 
-        extra_clause = ""
-        if allowed_statuses:
-            extra_clause = cur.mogrify("AND status IN %s",
-                                       (tuple(allowed_statuses),)).decode()
+        query_parts = [
+            'SELECT %s' % ', '.join(self.origin_visit_get_cols),
+            'FROM origin_visit']
 
-        query = """\
-            SELECT %s
-            FROM origin_visit
-            INNER JOIN snapshot ON (origin_visit.snapshot=snapshot.id)
-            WHERE
-                origin = %%s AND snapshot is not null %s
-            ORDER BY date DESC, visit DESC
-            LIMIT 1
-            """ % (', '.join(self.origin_visit_get_cols), extra_clause)
+        query_parts.append('WHERE origin = %s')
+
+        if require_snapshot:
+            query_parts.append('AND snapshot is not null')
+
+        if allowed_statuses:
+            query_parts.append(
+                cur.mogrify('AND status IN %s',
+                            (tuple(allowed_statuses),)).decode())
+
+        query_parts.append('ORDER BY date DESC, visit DESC LIMIT 1')
+
+        query = '\n'.join(query_parts)
 
         cur.execute(query, (origin_id,))
         r = cur.fetchone()
@@ -619,7 +667,8 @@ class Db(BaseDb):
 
         query = """SELECT %s FROM (VALUES %%s) as t(type, url)
                    LEFT JOIN origin
-                       ON (t.type=origin.type AND t.url=origin.url)
+                       ON ((t.type IS NULL OR t.type=origin.type)
+                           AND t.url=origin.url)
                 """ % ','.join('origin.' + col for col in self.origin_cols)
 
         yield from execute_values_generator(
@@ -739,12 +788,12 @@ class Db(BaseDb):
         """
         cur = self._cursor(cur)
 
-        query = """SELECT %s
-                   FROM person
-                   WHERE id IN %%s""" % ', '.join(self.person_get_cols)
+        query = """SELECT %s FROM (VALUES %%s) as t(id)
+                   LEFT JOIN person ON t.id = person.id
+                """ % ','.join('person.' + col for col in self.person_get_cols)
 
-        cur.execute(query, (tuple(ids),))
-        yield from cur
+        yield from execute_values_generator(
+            cur, query, ((id,) for id in ids))
 
     release_add_cols = [
         'id', 'target', 'target_type', 'date', 'date_offset',
@@ -806,7 +855,7 @@ class Db(BaseDb):
             query = '''SELECT %s
                        FROM swh_origin_metadata_get_by_origin(
                             %%s)''' % (','.join(
-                                          self.origin_metadata_get_cols))
+                self.origin_metadata_get_cols))
 
             cur.execute(query, (origin_id, ))
 
@@ -814,7 +863,7 @@ class Db(BaseDb):
             query = '''SELECT %s
                        FROM swh_origin_metadata_get_by_provider_type(
                             %%s, %%s)''' % (','.join(
-                                          self.origin_metadata_get_cols))
+                self.origin_metadata_get_cols))
 
             cur.execute(query, (origin_id, provider_type))
 
@@ -839,8 +888,8 @@ class Db(BaseDb):
                        where name=%%s and
                              version=%%s and
                              configuration=%%s''' % (
-                                 ','.join(self.tool_cols)),
-                    (name, version, configuration))
+            ','.join(self.tool_cols)),
+            (name, version, configuration))
 
         return cur.fetchone()
 
@@ -856,7 +905,7 @@ class Db(BaseDb):
                     RETURNING id"""
 
         cur.execute(insert, (provider_name, provider_type, provider_url,
-                    jsonize(metadata)))
+                             jsonize(metadata)))
         return cur.fetchone()[0]
 
     def metadata_provider_get(self, provider_id, cur=None):
@@ -864,8 +913,8 @@ class Db(BaseDb):
         cur.execute('''select %s
                        from metadata_provider
                        where id=%%s ''' % (
-                                 ','.join(self.metadata_provider_cols)),
-                    (provider_id, ))
+            ','.join(self.metadata_provider_cols)),
+            (provider_id, ))
 
         return cur.fetchone()
 
@@ -876,7 +925,7 @@ class Db(BaseDb):
                        from metadata_provider
                        where provider_name=%%s and
                              provider_url=%%s''' % (
-                                 ','.join(self.metadata_provider_cols)),
-                    (provider_name, provider_url))
+            ','.join(self.metadata_provider_cols)),
+            (provider_name, provider_url))
 
         return cur.fetchone()
