@@ -3,6 +3,7 @@
 # License: GNU General Public License version 3, or any later version
 # See top-level LICENSE file for more information
 
+import os
 import re
 import bisect
 import dateutil
@@ -29,17 +30,30 @@ def now():
     return datetime.datetime.now(tz=datetime.timezone.utc)
 
 
+ENABLE_ORIGIN_IDS = \
+    os.environ.get('SWH_STORAGE_IN_MEMORY_ENABLE_ORIGIN_IDS', 'true') == 'true'
+
+
 class Storage:
     def __init__(self, journal_writer=None):
         self._contents = {}
         self._content_indexes = defaultdict(lambda: defaultdict(set))
 
+        self.reset()
+
+        if journal_writer:
+            self.journal_writer = get_journal_writer(**journal_writer)
+        else:
+            self.journal_writer = None
+
+    def reset(self):
         self._directories = {}
         self._revisions = {}
         self._releases = {}
         self._snapshots = {}
-        self._origins = []
-        self._origin_visits = []
+        self._origins = {}
+        self._origins_by_id = []
+        self._origin_visits = {}
         self._persons = []
         self._origin_metadata = defaultdict(list)
         self._tools = {}
@@ -50,10 +64,6 @@ class Storage:
         self._sorted_sha1s = []
 
         self.objstorage = get_objstorage('memory', {})
-        if journal_writer:
-            self.journal_writer = get_journal_writer(**journal_writer)
-        else:
-            self.journal_writer = None
 
     def check_config(self, *, check_write):
         """Check that the storage is configured and ready to go."""
@@ -76,7 +86,8 @@ class Storage:
             if key in self._contents:
                 continue
             for algorithm in DEFAULT_ALGORITHMS:
-                if content[algorithm] in self._content_indexes[algorithm]:
+                if content[algorithm] in self._content_indexes[algorithm]\
+                   and (algorithm not in {'blake2s256', 'sha256'}):
                     from . import HashCollision
                     raise HashCollision(algorithm, content[algorithm], key)
             for algorithm in DEFAULT_ALGORITHMS:
@@ -99,11 +110,11 @@ class Storage:
         }
 
         if with_data:
-            summary['content:bytes:add'] = count_content_bytes_added
+            summary['content:add:bytes'] = count_content_bytes_added
 
         return summary
 
-    def content_add(self, contents):
+    def content_add(self, content):
         """Add content blobs to the storage
 
         Args:
@@ -132,13 +143,13 @@ class Storage:
                 skipped_content:add: New skipped contents (no data) added
 
         """
-        contents = [dict(c.items()) for c in contents]  # semi-shallow copy
+        content = [dict(c.items()) for c in content]  # semi-shallow copy
         now = datetime.datetime.now(tz=datetime.timezone.utc)
-        for item in contents:
+        for item in content:
             item['ctime'] = now
-        return self._content_add(contents, with_data=True)
+        return self._content_add(content, with_data=True)
 
-    def content_add_metadata(self, contents):
+    def content_add_metadata(self, content):
         """Add content metadata to the storage (like `content_add`, but
         without inserting to the objstorage).
 
@@ -167,9 +178,9 @@ class Storage:
                 skipped_content:add: New skipped contents (no data) added
 
         """
-        return self._content_add(contents, with_data=False)
+        return self._content_add(content, with_data=False)
 
-    def content_get(self, ids):
+    def content_get(self, content):
         """Retrieve in bulk contents and their data.
 
         This function may yield more blobs than provided sha1 identifiers,
@@ -191,10 +202,10 @@ class Storage:
 
         """
         # FIXME: Make this method support slicing the `data`.
-        if len(ids) > BULK_BLOCK_CONTENT_LEN_MAX:
+        if len(content) > BULK_BLOCK_CONTENT_LEN_MAX:
             raise ValueError(
                 "Sending at most %s contents." % BULK_BLOCK_CONTENT_LEN_MAX)
-        for obj_id in ids:
+        for obj_id in content:
             try:
                 data = self.objstorage.get(obj_id)
             except ObjNotFoundError:
@@ -247,7 +258,7 @@ class Storage:
             'next': next_content,
         }
 
-    def content_get_metadata(self, sha1s):
+    def content_get_metadata(self, content):
         """Retrieve content metadata in bulk
 
         Args:
@@ -258,7 +269,7 @@ class Storage:
         """
         # FIXME: the return value should be a mapping from search key to found
         # content*s*
-        for sha1 in sha1s:
+        for sha1 in content:
             if sha1 in self._content_indexes['sha1']:
                 objs = self._content_indexes['sha1'][sha1]
                 # FIXME: rather than selecting one of the objects with that
@@ -288,14 +299,14 @@ class Storage:
             hash = content.get(algo)
             if hash and hash in self._content_indexes[algo]:
                 found.append(self._content_indexes[algo][hash])
+
         if not found:
-            return
+            return []
+
         keys = list(set.intersection(*found))
+        return copy.deepcopy([self._contents[key] for key in keys])
 
-        # FIXME: should really be a list of all the objects found
-        return copy.deepcopy(self._contents[keys[0]])
-
-    def content_missing(self, contents, key_hash='sha1'):
+    def content_missing(self, content, key_hash='sha1'):
         """List content missing from storage
 
         Args:
@@ -312,18 +323,17 @@ class Storage:
             iterable ([bytes]): missing content ids (as per the
             key_hash column)
         """
-        for content in contents:
-            for (algo, hash_) in content.items():
+        for cont in content:
+            for (algo, hash_) in cont.items():
                 if algo not in DEFAULT_ALGORITHMS:
                     continue
                 if hash_ not in self._content_indexes.get(algo, []):
-                    yield content[key_hash]
+                    yield cont[key_hash]
                     break
             else:
-                # content_find cannot return None here, because we checked
-                # above that there is a content with matching hashes.
-                if self.content_find(content)['status'] == 'missing':
-                    yield content[key_hash]
+                for result in self.content_find(cont):
+                    if result['status'] == 'missing':
+                        yield cont[key_hash]
 
     def content_missing_per_sha1(self, contents):
         """List content missing from storage based only on sha1.
@@ -379,7 +389,7 @@ class Storage:
 
         return {'directory:add': count}
 
-    def directory_missing(self, directory_ids):
+    def directory_missing(self, directories):
         """List directories missing from storage
 
         Args:
@@ -389,7 +399,7 @@ class Storage:
             missing directory ids
 
         """
-        for id in directory_ids:
+        for id in directories:
             if id not in self._directories:
                 yield id
 
@@ -404,8 +414,10 @@ class Storage:
         ret = dict.fromkeys(keys)
         ret.update(dentry)
         if ret['type'] == 'file':
+            # TODO: Make it able to handle more than one content
             content = self.content_find({'sha1_git': ret['target']})
             if content:
+                content = content[0]
                 for key in keys:
                     ret[key] = content[key]
         return ret
@@ -421,7 +433,7 @@ class Storage:
                     yield from self._directory_ls(
                         ret['target'], True, prefix + ret['name'] + b'/')
 
-    def directory_ls(self, directory_id, recursive=False):
+    def directory_ls(self, directory, recursive=False):
         """Get entries for one directory.
 
         Args:
@@ -434,7 +446,7 @@ class Storage:
         If `recursive=True`, names in the path of a dir/file not at the
         root are concatenated with a slash (`/`).
         """
-        yield from self._directory_ls(directory_id, recursive)
+        yield from self._directory_ls(directory, recursive)
 
     def directory_entry_get_by_path(self, directory, paths):
         """Get the directory entry (either file or dir) from directory with path.
@@ -448,6 +460,9 @@ class Storage:
             The corresponding directory entry if found, None otherwise.
 
         """
+        return self._directory_entry_get_by_path(directory, paths, b'')
+
+    def _directory_entry_get_by_path(self, directory, paths, prefix):
         if not paths:
             return
 
@@ -459,6 +474,8 @@ class Storage:
         def _get_entry(entries, name):
             for entry in entries:
                 if entry['name'] == name:
+                    entry = entry.copy()
+                    entry['name'] = prefix + entry['name']
                     return entry
 
         first_item = _get_entry(contents, paths[0])
@@ -469,8 +486,8 @@ class Storage:
         if not first_item or first_item['type'] != 'dir':
             return
 
-        return self.directory_entry_get_by_path(
-                first_item['target'], paths[1:])
+        return self._directory_entry_get_by_path(
+                first_item['target'], paths[1:], prefix + paths[0] + b'/')
 
     def revision_add(self, revisions):
         """Add revisions to the storage
@@ -527,7 +544,7 @@ class Storage:
 
         return {'revision:add': count}
 
-    def revision_missing(self, revision_ids):
+    def revision_missing(self, revisions):
         """List revisions missing from storage
 
         Args:
@@ -537,12 +554,12 @@ class Storage:
             missing revision ids
 
         """
-        for id in revision_ids:
+        for id in revisions:
             if id not in self._revisions:
                 yield id
 
-    def revision_get(self, revision_ids):
-        for id in revision_ids:
+    def revision_get(self, revisions):
+        for id in revisions:
             yield copy.deepcopy(self._revisions.get(id))
 
     def _get_parent_revs(self, rev_id, seen, limit):
@@ -555,7 +572,7 @@ class Storage:
         for parent in self._revisions[rev_id]['parents']:
             yield from self._get_parent_revs(parent, seen, limit)
 
-    def revision_log(self, revision_ids, limit=None):
+    def revision_log(self, revisions, limit=None):
         """Fetch revision entry from the given root revisions.
 
         Args:
@@ -567,7 +584,7 @@ class Storage:
 
         """
         seen = set()
-        for rev_id in revision_ids:
+        for rev_id in revisions:
             yield from self._get_parent_revs(rev_id, seen, limit)
 
     def revision_shortlog(self, revisions, limit=None):
@@ -618,7 +635,8 @@ class Storage:
             if rel['id'] not in self._releases:
                 rel = copy.deepcopy(rel)
                 rel['date'] = normalize_timestamp(rel['date'])
-                self._person_add(rel['author'])
+                if rel['author']:
+                    self._person_add(rel['author'])
                 self._objects[rel['id']].append(
                     ('release', rel['id']))
                 self._releases[rel['id']] = rel
@@ -652,7 +670,7 @@ class Storage:
         for rel_id in releases:
             yield copy.deepcopy(self._releases.get(rel_id))
 
-    def snapshot_add(self, snapshots, legacy_arg1=None, legacy_arg2=None):
+    def snapshot_add(self, snapshots):
         """Add a snapshot to the storage
 
         Args:
@@ -681,13 +699,6 @@ class Storage:
                 snapshot_added: Count of object actually stored in db
 
         """
-        if legacy_arg1:
-            assert legacy_arg2
-            (origin, visit, snapshots) = \
-                (snapshots, legacy_arg1, [legacy_arg2])
-        else:
-            origin = visit = None
-
         count = 0
         for snapshot in snapshots:
             snapshot_id = snapshot['id']
@@ -703,11 +714,6 @@ class Storage:
                 self._objects[snapshot_id].append(('snapshot', snapshot_id))
                 count += 1
 
-        if origin:
-            # Legacy API, there can be only one snapshot
-            self.origin_visit_update(
-                origin, visit, snapshot=snapshots[0]['id'])
-
         return {'snapshot:add': count}
 
     def snapshot_get(self, snapshot_id):
@@ -759,10 +765,14 @@ class Storage:
                   branches.
 
         """
-        if origin > len(self._origins) or \
-           visit > len(self._origin_visits[origin-1]):
+        origin_url = self._get_origin_url(origin)
+        if not origin_url:
+            return
+
+        if origin_url not in self._origins or \
+           visit > len(self._origin_visits[origin_url]):
             return None
-        snapshot_id = self._origin_visits[origin-1][visit-1]['snapshot']
+        snapshot_id = self._origin_visits[origin_url][visit-1]['snapshot']
         if snapshot_id:
             return self.snapshot_get(snapshot_id)
         else:
@@ -778,13 +788,13 @@ class Storage:
 
         .. warning:: At most 1000 branches contained in the snapshot will be
             returned for performance reasons. In order to browse the whole
-            set of branches, the method :meth:`snapshot_get_branches`
-            should be used instead.
+            set of branches, the methods :meth:`origin_visit_get_latest`
+            and :meth:`snapshot_get_branches` should be used instead.
 
         Args:
-            origin (int): the origin's identifier
+            origin (Union[str,int]): the origin's URL or identifier
             allowed_statuses (list of str): list of visit statuses considered
-                to find the latest snapshot for the visit. For instance,
+                to find the latest snapshot for the origin. For instance,
                 ``allowed_statuses=['full']`` will only consider visits that
                 have successfully run to completion.
         Returns:
@@ -796,19 +806,20 @@ class Storage:
                   or :const:`None` if the snapshot has less than 1000
                   branches.
         """
-        visits = self._origin_visits[origin-1]
-        if allowed_statuses is not None:
-            visits = [visit for visit in visits
-                      if visit['status'] in allowed_statuses]
-        snapshot = None
-        for visit in sorted(visits, key=lambda v: (v['date'], v['visit']),
-                            reverse=True):
-            snapshot_id = visit['snapshot']
-            snapshot = self.snapshot_get(snapshot_id)
-            if snapshot:
-                break
+        origin_url = self._get_origin_url(origin)
+        if not origin_url:
+            return
 
-        return snapshot
+        visit = self.origin_visit_get_latest(
+            origin_url,
+            allowed_statuses=allowed_statuses,
+            require_snapshot=True)
+        if visit and visit['snapshot']:
+            snapshot = self.snapshot_get(visit['snapshot'])
+            if not snapshot:
+                raise ValueError(
+                    'last origin visit references an unknown snapshot')
+            return snapshot
 
     def snapshot_count_branches(self, snapshot_id, db=None, cur=None):
         """Count the number of branches in the snapshot with the given id
@@ -915,10 +926,13 @@ class Storage:
         """Return origins, either all identified by their ids or all
         identified by tuples (type, url).
 
+        If the url is given and the type is omitted, one of the origins with
+        that url is returned.
+
         Args:
             origin: a list of dictionaries representing the individual
                 origins to find.
-                These dicts have either the keys type and url:
+                These dicts have either the key url (and optionally type):
 
                 - type (FIXME: enum TBD): the origin type ('git', 'wget', ...)
                 - url (bytes): the url the origin points to
@@ -950,28 +964,26 @@ class Storage:
                 and not all('id' in origin for origin in origins):
             raise ValueError(
                 'Either all origins or none at all should have an "id".')
-        if any('type' in origin and 'url' in origin for origin in origins) \
-                and not all('type' in origin and 'url' in origin
-                            for origin in origins):
+        if any('url' in origin for origin in origins) \
+                and not all('url' in origin for origin in origins):
             raise ValueError(
-                'Either all origins or none at all should have a '
-                '"type" and an "url".')
+                'Either all origins or none at all should have '
+                'an "url" key.')
 
         results = []
         for origin in origins:
+            result = None
             if 'id' in origin:
-                origin_id = origin['id']
-            elif 'type' in origin and 'url' in origin:
-                origin_id = self._origin_id(origin)
+                assert ENABLE_ORIGIN_IDS, 'origin ids are disabled'
+                if origin['id'] <= len(self._origins_by_id):
+                    result = self._origins[self._origins_by_id[origin['id']-1]]
+            elif 'url' in origin:
+                if origin['url'] in self._origins:
+                    result = copy.deepcopy(self._origins[origin['url']])
             else:
                 raise ValueError(
-                    'Origin must have either id or (type and url).')
-            origin = None
-            # self._origin_id can return None
-            if origin_id is not None and origin_id <= len(self._origins):
-                origin = copy.deepcopy(self._origins[origin_id-1])
-                origin['id'] = origin_id
-            results.append(origin)
+                    'Origin must have either id or url.')
+            results.append(result)
 
         if return_single:
             assert len(results) == 1
@@ -994,12 +1006,12 @@ class Storage:
             by :meth:`swh.storage.in_memory.Storage.origin_get`.
         """
         origin_from = max(origin_from, 1)
-        if origin_from <= len(self._origins):
+        if origin_from <= len(self._origins_by_id):
             max_idx = origin_from + origin_count - 1
-            if max_idx > len(self._origins):
-                max_idx = len(self._origins)
+            if max_idx > len(self._origins_by_id):
+                max_idx = len(self._origins_by_id)
             for idx in range(origin_from-1, max_idx):
-                yield copy.deepcopy(self._origins[idx])
+                yield copy.deepcopy(self._origins[self._origins_by_id[idx]])
 
     def origin_search(self, url_pattern, offset=0, limit=50,
                       regexp=False, with_visit=False, db=None, cur=None):
@@ -1020,15 +1032,18 @@ class Storage:
             An iterable of dict containing origin information as returned
             by :meth:`swh.storage.storage.Storage.origin_get`.
         """
-        origins = self._origins
+        origins = self._origins.values()
         if regexp:
             pat = re.compile(url_pattern)
-            origins = [orig for orig in origins if pat.match(orig['url'])]
+            origins = [orig for orig in origins if pat.search(orig['url'])]
         else:
             origins = [orig for orig in origins if url_pattern in orig['url']]
         if with_visit:
             origins = [orig for orig in origins
-                       if len(self._origin_visits[orig['id']-1]) > 0]
+                       if len(self._origin_visits[orig['url']]) > 0]
+
+        if ENABLE_ORIGIN_IDS:
+            origins.sort(key=lambda origin: origin['id'])
 
         origins = copy.deepcopy(origins[offset:offset+limit])
         return origins
@@ -1069,7 +1084,10 @@ class Storage:
         """
         origins = copy.deepcopy(origins)
         for origin in origins:
-            origin['id'] = self.origin_add_one(origin)
+            if ENABLE_ORIGIN_IDS:
+                origin['id'] = self.origin_add_one(origin)
+            else:
+                self.origin_add_one(origin)
         return origins
 
     def origin_add_one(self, origin):
@@ -1089,26 +1107,32 @@ class Storage:
         """
         origin = copy.deepcopy(origin)
         assert 'id' not in origin
-        origin_id = self._origin_id(origin)
-        if origin_id is None:
+        if origin['url'] in self._origins:
+            if ENABLE_ORIGIN_IDS:
+                origin_id = self._origins[origin['url']]['id']
+        else:
             if self.journal_writer:
                 self.journal_writer.write_addition('origin', origin)
-            # origin ids are in the range [1, +inf[
-            origin_id = len(self._origins) + 1
-            origin['id'] = origin_id
-            self._origins.append(origin)
-            self._origin_visits.append([])
-            key = (origin['type'], origin['url'])
-            self._objects[key].append(('origin', origin_id))
-        else:
-            origin['id'] = origin_id
+            if ENABLE_ORIGIN_IDS:
+                # origin ids are in the range [1, +inf[
+                origin_id = len(self._origins) + 1
+                origin['id'] = origin_id
+                self._origins_by_id.append(origin['url'])
+                assert len(self._origins_by_id) == origin_id
+            self._origins[origin['url']] = origin
+            self._origin_visits[origin['url']] = []
+            self._objects[origin['url']].append(('origin', origin['url']))
 
-        return origin_id
+        if ENABLE_ORIGIN_IDS:
+            return origin_id
+        else:
+            return origin['url']
 
     def fetch_history_start(self, origin_id):
         """Add an entry for origin origin_id in fetch_history. Returns the id
         of the added fetch_history entry
         """
+        assert not ENABLE_ORIGIN_IDS, 'origin ids are disabled'
         pass
 
     def fetch_history_end(self, fetch_history_id, data):
@@ -1123,12 +1147,16 @@ class Storage:
         raise NotImplementedError('fetch_history_get is deprecated, use '
                                   'origin_visit_get instead.')
 
-    def origin_visit_add(self, origin, date=None, *, ts=None):
+    def origin_visit_add(self, origin, date=None, type=None, *, ts=None):
         """Add an origin_visit for the origin at date with status 'ongoing'.
 
+        For backward compatibility, `type` is optional and defaults to
+        the origin's type.
+
         Args:
-            origin (int): visited origin's identifier
+            origin (Union[int,str]): visited origin's identifier or URL
             date: timestamp of such visit
+            type (str): the type of loader used for the visit (hg, git, ...)
 
         Returns:
             dict: dictionary with keys origin and visit where:
@@ -1147,33 +1175,41 @@ class Storage:
                           DeprecationWarning)
             date = ts
 
-        origin_id = origin  # TODO: rename the argument
+        origin_url = self._get_origin_url(origin)
+        if origin_url is None:
+            raise ValueError('Unknown origin.')
 
         if isinstance(date, str):
             date = dateutil.parser.parse(date)
 
         visit_ret = None
-        if origin_id <= len(self._origin_visits):
+        if origin_url in self._origins:
+            origin = self._origins[origin_url]
             # visit ids are in the range [1, +inf[
-            visit_id = len(self._origin_visits[origin_id-1]) + 1
+            visit_id = len(self._origin_visits[origin_url]) + 1
             status = 'ongoing'
             visit = {
-                'origin': origin_id,
+                'origin': origin,
                 'date': date,
+                'type': type or origin['type'],
                 'status': status,
                 'snapshot': None,
                 'metadata': None,
                 'visit': visit_id
             }
-            self._origin_visits[origin_id-1].append(visit)
+            self._origin_visits[origin_url].append(visit)
             visit_ret = {
-                'origin': origin_id,
+                'origin': origin['id'] if ENABLE_ORIGIN_IDS else origin['url'],
                 'visit': visit_id,
             }
 
+            self._objects[(origin_url, visit_id)].append(
+                ('origin_visit', None))
+
             if self.journal_writer:
-                origin = self.origin_get([{'id': origin_id}])[0]
-                del origin['id']
+                origin = self._origins[origin_url].copy()
+                if 'id' in origin:
+                    del origin['id']
                 self.journal_writer.write_addition('origin_visit', {
                     **visit, 'origin': origin})
 
@@ -1184,7 +1220,7 @@ class Storage:
         """Update an origin_visit's status.
 
         Args:
-            origin (int): visited origin's identifier
+            origin (Union[int,str]): visited origin's identifier or URL
             visit_id (int): visit's identifier
             status: visit's new status
             metadata: data associated to the visit
@@ -1195,23 +1231,28 @@ class Storage:
             None
 
         """
-        origin_id = origin  # TODO: rename the argument
+        origin_url = self._get_origin_url(origin)
+        if origin_url is None:
+            raise ValueError('Unknown origin.')
 
         try:
-            visit = self._origin_visits[origin_id-1][visit_id-1]
+            visit = self._origin_visits[origin_url][visit_id-1]
         except IndexError:
-            raise ValueError('Invalid origin_id or visit_id') from None
+            raise ValueError('Unknown visit_id for this origin') \
+                from None
         if self.journal_writer:
-            origin = self.origin_get([{'id': origin_id}])[0]
-            del origin['id']
+            origin = self._origins[origin_url].copy()
+            if 'id' in origin:
+                del origin['id']
             self.journal_writer.write_update('origin_visit', {
-                'origin': origin, 'visit': visit_id,
+                'origin': origin, 'type': origin['type'],
+                'visit': visit_id,
                 'status': status or visit['status'],
                 'date': visit['date'],
                 'metadata': metadata or visit['metadata'],
                 'snapshot': snapshot or visit['snapshot']})
-        if origin_id > len(self._origin_visits) or \
-           visit_id > len(self._origin_visits[origin_id-1]):
+        if origin_url not in self._origin_visits or \
+                visit_id > len(self._origin_visits[origin_url]):
             return
         if status:
             visit['status'] = status
@@ -1220,6 +1261,63 @@ class Storage:
         if snapshot:
             visit['snapshot'] = snapshot
 
+    def origin_visit_upsert(self, visits):
+        """Add a origin_visits with a specific id and with all its data.
+        If there is already an origin_visit with the same
+        `(origin_url, visit_id)`, updates it instead of inserting a new one.
+
+        Args:
+            visits: iterable of dicts with keys:
+
+                origin: dict with keys either `id` or `url`
+                visit: origin visit id
+                type: type of loader used for the visit
+                date: timestamp of such visit
+                status: Visit's new status
+                metadata: Data associated to the visit
+                snapshot (sha1_git): identifier of the snapshot to add to
+                    the visit
+        """
+        visits = copy.deepcopy(visits)
+        for visit in visits:
+            if isinstance(visit['date'], str):
+                visit['date'] = dateutil.parser.parse(visit['date'])
+
+        if self.journal_writer:
+            for visit in visits:
+                visit = visit.copy()
+                visit['origin'] = self._origins[visit['origin']['url']].copy()
+                if 'id' in visit['origin']:
+                    del visit['origin']['id']
+                self.journal_writer.write_addition('origin_visit', visit)
+
+        for visit in visits:
+            visit_id = visit['visit']
+            origin_url = visit['origin']['url']
+
+            self._objects[(origin_url, visit_id)].append(
+                ('origin_visit', None))
+
+            while len(self._origin_visits[origin_url]) < visit_id:
+                self._origin_visits[origin_url].append(None)
+
+            visit = visit.copy()
+
+            visit = self._origin_visits[origin_url][visit_id-1] = visit
+
+    def _convert_visit(self, visit):
+        if visit is None:
+            return
+
+        visit = visit.copy()
+        origin = self._origins[visit['origin']['url']]
+        if ENABLE_ORIGIN_IDS:
+            visit['origin'] = origin['id']
+        else:
+            visit['origin'] = origin['url']
+
+        return visit
+
     def origin_visit_get(self, origin, last_visit=None, limit=None):
         """Retrieve all the origin's visit's information.
 
@@ -1234,15 +1332,41 @@ class Storage:
             List of visits.
 
         """
-        if origin <= len(self._origin_visits):
-            visits = self._origin_visits[origin-1]
+        origin_url = self._get_origin_url(origin)
+        if origin_url in self._origin_visits:
+            visits = self._origin_visits[origin_url]
             if last_visit is not None:
                 visits = visits[last_visit:]
             if limit is not None:
                 visits = visits[:limit]
             for visit in visits:
+                if not visit:
+                    continue
                 visit_id = visit['visit']
-                yield copy.deepcopy(self._origin_visits[origin-1][visit_id-1])
+
+                yield self._convert_visit(
+                    self._origin_visits[origin_url][visit_id-1])
+
+    def origin_visit_find_by_date(self, origin, visit_date):
+        """Retrieves the origin visit whose date is closest to the provided
+        timestamp.
+        In case of a tie, the visit with largest id is selected.
+
+        Args:
+            origin (str): The occurrence's origin (URL).
+            target (datetime): target timestamp
+
+        Returns:
+            A visit.
+
+        """
+        origin_url = self._get_origin_url(origin)
+        if origin_url in self._origin_visits:
+            visits = self._origin_visits[origin_url]
+            visit = min(
+                visits,
+                key=lambda v: (abs(v['date'] - visit_date), -v['visit']))
+            return self._convert_visit(visit)
 
     def origin_visit_get_by(self, origin, visit):
         """Retrieve origin visit's information.
@@ -1255,11 +1379,52 @@ class Storage:
             it does not exist
 
         """
-        origin_visit = None
-        if origin <= len(self._origin_visits) and \
-           visit <= len(self._origin_visits[origin-1]):
-            origin_visit = self._origin_visits[origin-1][visit-1]
-        return copy.deepcopy(origin_visit)
+        origin_url = self._get_origin_url(origin)
+        if origin_url in self._origin_visits and \
+           visit <= len(self._origin_visits[origin_url]):
+            return self._convert_visit(
+                self._origin_visits[origin_url][visit-1])
+
+    def origin_visit_get_latest(
+            self, origin, allowed_statuses=None, require_snapshot=False):
+        """Get the latest origin visit for the given origin, optionally
+        looking only for those with one of the given allowed_statuses
+        or for those with a known snapshot.
+
+        Args:
+            origin (str): the origin's URL
+            allowed_statuses (list of str): list of visit statuses considered
+                to find the latest visit. For instance,
+                ``allowed_statuses=['full']`` will only consider visits that
+                have successfully run to completion.
+            require_snapshot (bool): If True, only a visit with a snapshot
+                will be returned.
+        Returns:
+            dict: a dict with the following keys:
+
+                origin: the URL of the origin
+                visit: origin visit id
+                type: type of loader used for the visit
+                date: timestamp of such visit
+                status: Visit's new status
+                metadata: Data associated to the visit
+                snapshot (Optional[sha1_git]): identifier of the snapshot
+                    associated to the visit
+        """
+        origin = self._origins.get(origin)
+        if not origin:
+            return
+        visits = self._origin_visits[origin['url']]
+        if allowed_statuses is not None:
+            visits = [visit for visit in visits
+                      if visit['status'] in allowed_statuses]
+        if require_snapshot:
+            visits = [visit for visit in visits
+                      if visit['snapshot']]
+
+        visit = max(
+            visits, key=lambda v: (v['date'], v['visit']), default=None)
+        return self._convert_visit(visit)
 
     def person_get(self, person):
         """Return the persons identified by their ids.
@@ -1358,7 +1523,7 @@ class Storage:
             item = copy.deepcopy(item)
             provider = self.metadata_provider_get(item['provider_id'])
             for attr in ('name', 'type', 'url'):
-                item['provider_' + attr] = provider[attr]
+                item['provider_' + attr] = provider['provider_' + attr]
             metadata.append(item)
         return metadata
 
@@ -1420,9 +1585,9 @@ class Storage:
             an identifier of the provider
         """
         provider = {
-                'name': provider_name,
-                'type': provider_type,
-                'url': provider_url,
+                'provider_name': provider_name,
+                'provider_type': provider_type,
+                'provider_url': provider_url,
                 'metadata': metadata,
                 }
         key = self._metadata_provider_key(provider)
@@ -1453,19 +1618,19 @@ class Storage:
             dict: same as `metadata_provider_add`;
                   or None if it does not exist.
         """
-        key = self._metadata_provider_key({
-            'name': provider['provider_name'],
-            'url': provider['provider_url']})
+        key = self._metadata_provider_key(provider)
         return self._metadata_providers.get(key)
 
-    def _origin_id(self, origin):
-        origin_id = None
-        for stored_origin in self._origins:
-            if stored_origin['type'] == origin['type'] and \
-               stored_origin['url'] == origin['url']:
-                origin_id = stored_origin['id']
-                break
-        return origin_id
+    def _get_origin_url(self, origin):
+        if isinstance(origin, str):
+            return origin
+        elif isinstance(origin, int):
+            if origin <= len(self._origins_by_id):
+                return self._origins_by_id[origin-1]
+            else:
+                return None
+        else:
+            raise TypeError('origin must be a string or an integer.')
 
     def _person_add(self, person):
         """Add a person in storage.
@@ -1494,9 +1659,9 @@ class Storage:
 
     @staticmethod
     def _tool_key(tool):
-        return (tool['name'], tool['version'],
-                tuple(sorted(tool['configuration'].items())))
+        return '%r %r %r' % (tool['name'], tool['version'],
+                             tuple(sorted(tool['configuration'].items())))
 
     @staticmethod
     def _metadata_provider_key(provider):
-        return (provider['name'], provider['url'])
+        return '%r %r' % (provider['provider_name'], provider['provider_url'])
