@@ -623,7 +623,7 @@ begin
     union
         select committer_fullname as fullname, committer_name as name, committer_email as email from tmp_revision
     ) insert into person (fullname, name, email)
-    select distinct fullname, name, email from t
+    select distinct on (fullname) fullname, name, email from t
     where not exists (
         select 1
         from person p
@@ -660,8 +660,9 @@ as $$
 begin
     with t as (
         select distinct author_fullname as fullname, author_name as name, author_email as email from tmp_release
+        where author_fullname is not null
     ) insert into person (fullname, name, email)
-    select fullname, name, email from t
+    select distinct on (fullname) fullname, name, email from t
     where not exists (
         select 1
         from person p
@@ -692,7 +693,7 @@ $$;
 -- add a new origin_visit for origin origin_id at date.
 --
 -- Returns the new visit id.
-create or replace function swh_origin_visit_add(origin_id bigint, date timestamptz)
+create or replace function swh_origin_visit_add(origin_id bigint, date timestamptz, type text)
     returns bigint
     language sql
 as $$
@@ -701,8 +702,8 @@ as $$
     from origin_visit
     where origin = origin_id
   )
-  insert into origin_visit (origin, date, visit, status)
-  values (origin_id, date, (select visit from last_known_visit) + 1, 'ongoing')
+  insert into origin_visit (origin, date, type, visit, status)
+  values (origin_id, date, type, (select visit from last_known_visit) + 1, 'ongoing')
   returning visit;
 $$;
 
@@ -828,26 +829,27 @@ as $$
 $$;
 
 -- Find the visit of origin id closest to date visit_date
+-- Breaks ties by selecting the largest visit id
 create or replace function swh_visit_find_by_date(origin bigint, visit_date timestamptz default NOW())
     returns origin_visit
     language sql
     stable
 as $$
   with closest_two_visits as ((
-    select ov, (date - visit_date) as interval
+    select ov, (date - visit_date), visit as interval
     from origin_visit ov
     where ov.origin = origin
           and ov.date >= visit_date
-    order by ov.date asc
+    order by ov.date asc, ov.visit desc
     limit 1
   ) union (
-    select ov, (visit_date - date) as interval
+    select ov, (visit_date - date), visit as interval
     from origin_visit ov
     where ov.origin = origin
           and ov.date < visit_date
-    order by ov.date desc
+    order by ov.date desc, ov.visit desc
     limit 1
-  )) select (ov).* from closest_two_visits order by interval limit 1
+  )) select (ov).* from closest_two_visits order by interval, visit limit 1
 $$;
 
 -- Find the visit of origin id closest to date visit_date
