@@ -72,38 +72,40 @@ class Storage():
         else:
             return Db.from_pool(self._pool)
 
-    def check_config(self, *, check_write):
+    def put_db(self, db):
+        if db is not self._db:
+            db.put_conn()
+
+    @db_transaction()
+    def check_config(self, *, check_write, db, cur):
         """Check that the storage is configured and ready to go."""
 
         if not self.objstorage.check_config(check_write=check_write):
             return False
 
         # Check permissions on one of the tables
-        with self.get_db().transaction() as cur:
-            if check_write:
-                check = 'INSERT'
-            else:
-                check = 'SELECT'
+        if check_write:
+            check = 'INSERT'
+        else:
+            check = 'SELECT'
 
-            cur.execute(
-                "select has_table_privilege(current_user, 'content', %s)",
-                (check,)
-            )
-            return cur.fetchone()[0]
+        cur.execute(
+            "select has_table_privilege(current_user, 'content', %s)",
+            (check,)
+        )
+        return cur.fetchone()[0]
 
-        return True
-
-    def _content_unique_key(self, hash):
+    def _content_unique_key(self, hash, db):
         """Given a hash (tuple or dict), return a unique key from the
            aggregation of keys.
 
         """
-        keys = self.get_db().content_hash_keys
+        keys = db.content_hash_keys
         if isinstance(hash, tuple):
             return hash
         return tuple([hash[k] for k in keys])
 
-    def _filter_new_content(self, content):
+    def _filter_new_content(self, content, db, cur):
         content_by_status = defaultdict(list)
         for d in content:
             if 'status' not in d:
@@ -115,17 +117,18 @@ class Storage():
         content_with_data = content_by_status['visible']
         content_without_data = content_by_status['absent']
 
-        missing_content = set(self.content_missing(content_with_data))
-        missing_skipped = set(self._content_unique_key(hashes) for hashes
-                              in self.skipped_content_missing(
-                                  content_without_data))
+        missing_content = set(self.content_missing(content_with_data,
+                                                   db=db, cur=cur))
+        missing_skipped = set(self._content_unique_key(hashes, db)
+                              for hashes in self.skipped_content_missing(
+                                  content_without_data, db=db, cur=cur))
 
         content_with_data = [
             cont for cont in content_with_data
             if cont['sha1'] in missing_content]
         content_without_data = [
             cont for cont in content_without_data
-            if self._content_unique_key(cont) in missing_skipped]
+            if self._content_unique_key(cont, db) in missing_skipped]
 
         summary = {
             'content:add': len(missing_content),
@@ -169,7 +172,8 @@ class Storage():
             # move metadata in place
             db.skipped_content_add_from_temp(cur)
 
-    def content_add(self, content):
+    @db_transaction()
+    def content_add(self, content, db, cur):
         """Add content blobs to the storage
 
         Note: in case of DB errors, objects might have already been added to
@@ -204,7 +208,7 @@ class Storage():
             Summary dict with the following key and associated values:
 
                 content:add: New contents added
-                content:bytes:add: Sum of the contents' length data
+                content:add:bytes: Sum of the contents' length data
                 skipped_content:add: New skipped contents (no data) added
         """
         content = [dict(c.items()) for c in content]  # semi-shallow copy
@@ -219,10 +223,8 @@ class Storage():
                     del item['data']
                 self.journal_writer.write_addition('content', item)
 
-        db = self.get_db()
-
         (content_with_data, content_without_data, summary) = \
-            self._filter_new_content(content)
+            self._filter_new_content(content, db, cur)
 
         def add_to_objstorage():
             """Add to objstorage the new missing_content
@@ -245,18 +247,17 @@ class Storage():
             self.objstorage.add_batch(data)
             return content_bytes_added
 
-        with db.transaction() as cur:
-            with ThreadPoolExecutor(max_workers=1) as executor:
-                added_to_objstorage = executor.submit(add_to_objstorage)
+        with ThreadPoolExecutor(max_workers=1) as executor:
+            added_to_objstorage = executor.submit(add_to_objstorage)
 
-                self._content_add_metadata(
-                    db, cur, content_with_data, content_without_data)
+            self._content_add_metadata(
+                db, cur, content_with_data, content_without_data)
 
-                # Wait for objstorage addition before returning from the
-                # transaction, bubbling up any exception
-                content_bytes_added = added_to_objstorage.result()
+            # Wait for objstorage addition before returning from the
+            # transaction, bubbling up any exception
+            content_bytes_added = added_to_objstorage.result()
 
-        summary['content:bytes:add'] = content_bytes_added
+        summary['content:add:bytes'] = content_bytes_added
         return summary
 
     @db_transaction()
@@ -293,12 +294,13 @@ class Storage():
         db.content_update_from_temp(keys_to_update=keys,
                                     cur=cur)
 
-    def content_add_metadata(self, content):
+    @db_transaction()
+    def content_add_metadata(self, content, db, cur):
         """Add content metadata to the storage (like `content_add`, but
         without inserting to the objstorage).
 
         Args:
-            contents (iterable): iterable of dictionaries representing
+            content (iterable): iterable of dictionaries representing
                 individual pieces of content to add. Each dictionary has the
                 following keys:
 
@@ -323,14 +325,11 @@ class Storage():
                 assert 'data' not in content
                 self.journal_writer.write_addition('content', item)
 
-        db = self.get_db()
-
         (content_with_data, content_without_data, summary) = \
-            self._filter_new_content(content)
+            self._filter_new_content(content, db, cur)
 
-        with db.transaction() as cur:
-            self._content_add_metadata(
-                db, cur, content_with_data, content_without_data)
+        self._content_add_metadata(
+            db, cur, content_with_data, content_without_data)
 
         return summary
 
@@ -476,7 +475,7 @@ class Storage():
             yield obj[0]
 
     @db_transaction_generator()
-    def skipped_content_missing(self, content, db=None, cur=None):
+    def skipped_content_missing(self, contents, db=None, cur=None):
         """List skipped_content missing from storage
 
         Args:
@@ -487,13 +486,8 @@ class Storage():
             iterable: missing signatures
 
         """
-        keys = db.content_hash_keys
-
-        db.mktemp('skipped_content', cur)
-        db.copy_to(content, 'tmp_skipped_content',
-                   keys + ['length', 'reason'], cur)
-
-        yield from db.skipped_content_missing_from_temp(cur)
+        for content in db.skipped_content_missing(contents, cur):
+            yield dict(zip(db.content_hash_keys, content))
 
     @db_transaction()
     def content_find(self, content, db=None, cur=None):
@@ -517,16 +511,16 @@ class Storage():
             raise ValueError('content keys must contain at least one of: '
                              'sha1, sha1_git, sha256, blake2s256')
 
-        c = db.content_find(sha1=content.get('sha1'),
-                            sha1_git=content.get('sha1_git'),
-                            sha256=content.get('sha256'),
-                            blake2s256=content.get('blake2s256'),
-                            cur=cur)
-        if c:
-            return dict(zip(db.content_find_cols, c))
-        return None
+        contents = db.content_find(sha1=content.get('sha1'),
+                                   sha1_git=content.get('sha1_git'),
+                                   sha256=content.get('sha256'),
+                                   blake2s256=content.get('blake2s256'),
+                                   cur=cur)
+        return [dict(zip(db.content_find_cols, content))
+                for content in contents]
 
-    def directory_add(self, directories):
+    @db_transaction()
+    def directory_add(self, directories, db, cur):
         """Add directories to the storage
 
         Args:
@@ -570,37 +564,35 @@ class Storage():
                 entry['dir_id'] = dir_id
                 dir_entries[entry['type']][dir_id].append(entry)
 
-        dirs_missing = set(self.directory_missing(dirs))
+        dirs_missing = set(self.directory_missing(dirs, db=db, cur=cur))
         if not dirs_missing:
             return summary
 
-        db = self.get_db()
-        with db.transaction() as cur:
-            # Copy directory ids
-            dirs_missing_dict = ({'id': dir} for dir in dirs_missing)
-            db.mktemp('directory', cur)
-            db.copy_to(dirs_missing_dict, 'tmp_directory', ['id'], cur)
+        # Copy directory ids
+        dirs_missing_dict = ({'id': dir} for dir in dirs_missing)
+        db.mktemp('directory', cur)
+        db.copy_to(dirs_missing_dict, 'tmp_directory', ['id'], cur)
 
-            # Copy entries
-            for entry_type, entry_list in dir_entries.items():
-                entries = itertools.chain.from_iterable(
-                    entries_for_dir
-                    for dir_id, entries_for_dir
-                    in entry_list.items()
-                    if dir_id in dirs_missing)
+        # Copy entries
+        for entry_type, entry_list in dir_entries.items():
+            entries = itertools.chain.from_iterable(
+                entries_for_dir
+                for dir_id, entries_for_dir
+                in entry_list.items()
+                if dir_id in dirs_missing)
 
-                db.mktemp_dir_entry(entry_type)
+            db.mktemp_dir_entry(entry_type)
 
-                db.copy_to(
-                    entries,
-                    'tmp_directory_entry_%s' % entry_type,
-                    ['target', 'name', 'perms', 'dir_id'],
-                    cur,
-                )
+            db.copy_to(
+                entries,
+                'tmp_directory_entry_%s' % entry_type,
+                ['target', 'name', 'perms', 'dir_id'],
+                cur,
+            )
 
-            # Do the final copy
-            db.directory_add_from_temp(cur)
-            summary['directory:add'] = len(dirs_missing)
+        # Do the final copy
+        db.directory_add_from_temp(cur)
+        summary['directory:add'] = len(dirs_missing)
 
         return summary
 
@@ -658,7 +650,8 @@ class Storage():
         if res:
             return dict(zip(db.directory_ls_cols, res))
 
-    def revision_add(self, revisions):
+    @db_transaction()
+    def revision_add(self, revisions, db, cur):
         """Add revisions to the storage
 
         Args:
@@ -700,32 +693,30 @@ class Storage():
         if self.journal_writer:
             self.journal_writer.write_additions('revision', revisions)
 
-        db = self.get_db()
-
         revisions_missing = set(self.revision_missing(
-            set(revision['id'] for revision in revisions)))
+            set(revision['id'] for revision in revisions),
+            db=db, cur=cur))
 
         if not revisions_missing:
             return summary
 
-        with db.transaction() as cur:
-            db.mktemp_revision(cur)
+        db.mktemp_revision(cur)
 
-            revisions_filtered = (
-                converters.revision_to_db(revision) for revision in revisions
-                if revision['id'] in revisions_missing)
+        revisions_filtered = (
+            converters.revision_to_db(revision) for revision in revisions
+            if revision['id'] in revisions_missing)
 
-            parents_filtered = []
+        parents_filtered = []
 
-            db.copy_to(
-                revisions_filtered, 'tmp_revision', db.revision_add_cols,
-                cur,
-                lambda rev: parents_filtered.extend(rev['parents']))
+        db.copy_to(
+            revisions_filtered, 'tmp_revision', db.revision_add_cols,
+            cur,
+            lambda rev: parents_filtered.extend(rev['parents']))
 
-            db.revision_add_from_temp(cur)
+        db.revision_add_from_temp(cur)
 
-            db.copy_to(parents_filtered, 'revision_history',
-                       ['id', 'parent_id', 'parent_rank'], cur)
+        db.copy_to(parents_filtered, 'revision_history',
+                   ['id', 'parent_id', 'parent_rank'], cur)
 
         return {'revision:add': len(revisions_missing)}
 
@@ -803,7 +794,8 @@ class Storage():
 
         yield from db.revision_shortlog(revisions, limit, cur)
 
-    def release_add(self, releases):
+    @db_transaction()
+    def release_add(self, releases, db, cur):
         """Add releases to the storage
 
         Args:
@@ -834,26 +826,26 @@ class Storage():
         if self.journal_writer:
             self.journal_writer.write_additions('release', releases)
 
-        db = self.get_db()
-
         release_ids = set(release['id'] for release in releases)
-        releases_missing = set(self.release_missing(release_ids))
+        releases_missing = set(self.release_missing(release_ids,
+                                                    db=db, cur=cur))
 
         if not releases_missing:
             return summary
 
-        with db.transaction() as cur:
-            db.mktemp_release(cur)
+        db.mktemp_release(cur)
 
-            releases_filtered = (
-                converters.release_to_db(release) for release in releases
-                if release['id'] in releases_missing
-            )
+        releases_missing = list(releases_missing)
 
-            db.copy_to(releases_filtered, 'tmp_release', db.release_add_cols,
-                       cur)
+        releases_filtered = (
+            converters.release_to_db(release) for release in releases
+            if release['id'] in releases_missing
+        )
 
-            db.release_add_from_temp(cur)
+        db.copy_to(releases_filtered, 'tmp_release', db.release_add_cols,
+                   cur)
+
+        db.release_add_from_temp(cur)
 
         return {'release:add': len(releases_missing)}
 
@@ -893,8 +885,7 @@ class Storage():
             yield data if data['target_type'] else None
 
     @db_transaction()
-    def snapshot_add(self, snapshots, origin=None, visit=None,
-                     db=None, cur=None):
+    def snapshot_add(self, snapshots, db=None, cur=None):
         """Add snapshots to the storage.
 
         Args:
@@ -913,8 +904,6 @@ class Storage():
                 - **target** (:class:`bytes`): identifier of the target
                   (currently a ``sha1_git`` for all object kinds, or the name
                   of the target branch for aliases)
-            origin (int): legacy argument for backward compatibility
-            visit (int): legacy argument for backward compatibility
 
         Raises:
             ValueError: if the origin or visit id does not exist.
@@ -926,24 +915,6 @@ class Storage():
                 snapshot:add: Count of object actually stored in db
 
         """
-        if origin:
-            if not visit:
-                raise TypeError(
-                    'snapshot_add expects one argument (or, as a legacy '
-                    'behavior, three arguments), not two')
-            if isinstance(snapshots, int):
-                # Called by legacy code that uses the new api/client.py
-                (origin_id, visit_id, snapshots) = \
-                    (snapshots, origin, [visit])
-            else:
-                # Called by legacy code that uses the old api/client.py
-                origin_id = origin
-                visit_id = visit
-                snapshots = [snapshots]
-        else:
-            # Called by new code that uses the new api/client.py
-            origin_id = visit_id = None
-
         created_temp_table = False
 
         count = 0
@@ -974,12 +945,6 @@ class Storage():
                 db.snapshot_add(snapshot['id'], cur)
                 count += 1
 
-        if visit_id:
-            # Legacy API, there can be only one snapshot
-            self.origin_visit_update(
-                origin_id, visit_id, snapshot=snapshots[0]['id'],
-                db=db, cur=cur)
-
         return {'snapshot:add': count}
 
     @db_transaction(statement_timeout=2000)
@@ -1057,7 +1022,7 @@ class Storage():
             should be used instead.
 
         Args:
-            origin (int): the origin identifier
+            origin (Union[str,int]): the origin's URL or identifier
             allowed_statuses (list of str): list of visit statuses considered
                 to find the latest snapshot for the visit. For instance,
                 ``allowed_statuses=['full']`` will only consider visits that
@@ -1071,11 +1036,22 @@ class Storage():
                   or :const:`None` if the snapshot has less than 1000
                   branches.
         """
-        origin_visit = db.origin_visit_get_latest_snapshot(
-            origin, allowed_statuses=allowed_statuses, cur=cur)
-        if origin_visit:
-            origin_visit = dict(zip(db.origin_visit_get_cols, origin_visit))
-            return self.snapshot_get(origin_visit['snapshot'], db=db, cur=cur)
+        if isinstance(origin, int):
+            origin = self.origin_get({'id': origin}, db=db, cur=cur)
+            if not origin:
+                return
+            origin = origin['url']
+
+        origin_visit = self.origin_visit_get_latest(
+            origin, allowed_statuses=allowed_statuses, require_snapshot=True,
+            db=db, cur=cur)
+        if origin_visit and origin_visit['snapshot']:
+            snapshot = self.snapshot_get(
+                    origin_visit['snapshot'], db=db, cur=cur)
+            if not snapshot:
+                raise ValueError(
+                    'last origin visit references an unknown snapshot')
+            return snapshot
 
     @db_transaction(statement_timeout=2000)
     def snapshot_count_branches(self, snapshot_id, db=None, cur=None):
@@ -1157,13 +1133,17 @@ class Storage():
         return None
 
     @db_transaction()
-    def origin_visit_add(self, origin, date=None, db=None, cur=None, *,
-                         ts=None):
+    def origin_visit_add(self, origin, date=None, type=None,
+                         db=None, cur=None, *, ts=None):
         """Add an origin_visit for the origin at ts with status 'ongoing'.
 
+        For backward compatibility, `type` is optional and defaults to
+        the origin's type.
+
         Args:
-            origin: Visited Origin id
+            origin (Union[int,str]): visited origin's identifier or URL
             date: timestamp of such visit
+            type (str): the type of loader used for the visit (hg, git, ...)
 
         Returns:
             dict: dictionary with keys origin and visit where:
@@ -1182,20 +1162,28 @@ class Storage():
                           DeprecationWarning)
             date = ts
 
-        origin_id = origin  # TODO: rename the argument
+        if isinstance(origin, str):
+            origin = self.origin_get({'url': origin}, db=db, cur=cur)
+            origin_id = origin['id']
+        else:
+            origin = self.origin_get({'id': origin}, db=db, cur=cur)
+            origin_id = origin['id']
 
         if isinstance(date, str):
             date = dateutil.parser.parse(date)
 
-        visit_id = db.origin_visit_add(origin_id, date, cur)
+        if type is None:
+            type = origin['type']
+
+        visit_id = db.origin_visit_add(origin_id, date, type, cur)
 
         if self.journal_writer:
             # We can write to the journal only after inserting to the
             # DB, because we want the id of the visit
-            origin = self.origin_get([{'id': origin_id}], db=db, cur=cur)[0]
             del origin['id']
             self.journal_writer.write_addition('origin_visit', {
-                'origin': origin, 'date': date, 'visit': visit_id,
+                'origin': origin, 'date': date, 'type': type,
+                'visit': visit_id,
                 'status': 'ongoing', 'metadata': None, 'snapshot': None})
 
         return {
@@ -1210,7 +1198,7 @@ class Storage():
         """Update an origin_visit's status.
 
         Args:
-            origin: Visited Origin id
+            origin (Union[int,str]): visited origin's identifier or URL
             visit_id: Visit's id
             status: Visit's new status
             metadata: Data associated to the visit
@@ -1221,7 +1209,10 @@ class Storage():
             None
 
         """
-        origin_id = origin  # TODO: rename the argument
+        if isinstance(origin, str):
+            origin_id = self.origin_get({'url': origin}, db=db, cur=cur)['id']
+        else:
+            origin_id = origin
 
         visit = db.origin_visit_get(origin_id, visit_id, cur=cur)
 
@@ -1248,13 +1239,50 @@ class Storage():
 
             db.origin_visit_update(origin_id, visit_id, updates, cur)
 
+    @db_transaction()
+    def origin_visit_upsert(self, visits, db=None, cur=None):
+        """Add a origin_visits with a specific id and with all its data.
+        If there is already an origin_visit with the same
+        `(origin_id, visit_id)`, overwrites it.
+
+        Args:
+            visits: iterable of dicts with keys:
+
+                origin: dict with keys either `id` or `url`
+                visit: origin visit id
+                date: timestamp of such visit
+                status: Visit's new status
+                metadata: Data associated to the visit
+                snapshot (sha1_git): identifier of the snapshot to add to
+                    the visit
+        """
+        visits = copy.deepcopy(visits)
+        for visit in visits:
+            if isinstance(visit['date'], str):
+                visit['date'] = dateutil.parser.parse(visit['date'])
+            visit['origin'] = \
+                self.origin_get([visit['origin']], db=db, cur=cur)[0]
+
+        if self.journal_writer:
+            for visit in visits:
+                visit = copy.deepcopy(visit)
+                if visit.get('type') is None:
+                    visit['type'] = visit['origin']['type']
+                del visit['origin']['id']
+                self.journal_writer.write_addition('origin_visit', visit)
+
+        for visit in visits:
+            visit['origin'] = visit['origin']['id']
+            # TODO: upsert them all in a single query
+            db.origin_visit_upsert(**visit, cur=cur)
+
     @db_transaction_generator(statement_timeout=500)
     def origin_visit_get(self, origin, last_visit=None, limit=None, db=None,
                          cur=None):
         """Retrieve all the origin's visit's information.
 
         Args:
-            origin (int): The occurrence's origin (identifier).
+            origin (Union[int,str]): The occurrence's origin (identifier/URL).
             last_visit: Starting point from which listing the next visits
                 Default to None
             limit (int): Number of results to return from the last visit.
@@ -1264,11 +1292,38 @@ class Storage():
             List of visits.
 
         """
+        if isinstance(origin, str):
+            origin = self.origin_get([{'url': origin}], db=db, cur=cur)[0]
+            if not origin:
+                return
+            origin = origin['id']
         for line in db.origin_visit_get_all(
                 origin, last_visit=last_visit, limit=limit, cur=cur):
             data = dict(zip(db.origin_visit_get_cols, line))
             yield data
 
+    @db_transaction(statement_timeout=500)
+    def origin_visit_find_by_date(self, origin, visit_date, db=None, cur=None):
+        """Retrieves the origin visit whose date is closest to the provided
+        timestamp.
+        In case of a tie, the visit with largest id is selected.
+
+        Args:
+            origin (str): The occurrence's origin (URL).
+            target (datetime): target timestamp
+
+        Returns:
+            A visit.
+
+        """
+        origin = self.origin_get([{'url': origin}], db=db, cur=cur)[0]
+        if not origin:
+            return
+        origin = origin['id']
+        line = db.origin_visit_find_by_date(origin, visit_date, cur=cur)
+        if line:
+            return dict(zip(db.origin_visit_get_cols, line))
+
     @db_transaction(statement_timeout=500)
     def origin_visit_get_by(self, origin, visit, db=None, cur=None):
         """Retrieve origin visit's information.
@@ -1281,12 +1336,56 @@ class Storage():
             it does not exist
 
         """
+        if isinstance(origin, str):
+            origin = self.origin_get({'url': origin}, db=db, cur=cur)
+            if not origin:
+                return
+            origin = origin['id']
         ori_visit = db.origin_visit_get(origin, visit, cur)
         if not ori_visit:
             return None
 
         return dict(zip(db.origin_visit_get_cols, ori_visit))
 
+    @db_transaction(statement_timeout=4000)
+    def origin_visit_get_latest(
+            self, origin, allowed_statuses=None, require_snapshot=False,
+            db=None, cur=None):
+        """Get the latest origin visit for the given origin, optionally
+        looking only for those with one of the given allowed_statuses
+        or for those with a known snapshot.
+
+        Args:
+            origin (str): the origin's URL
+            allowed_statuses (list of str): list of visit statuses considered
+                to find the latest visit. For instance,
+                ``allowed_statuses=['full']`` will only consider visits that
+                have successfully run to completion.
+            require_snapshot (bool): If True, only a visit with a snapshot
+                will be returned.
+        Returns:
+            dict: a dict with the following keys:
+
+                origin: the URL of the origin
+                visit: origin visit id
+                type: type of loader used for the visit
+                date: timestamp of such visit
+                status: Visit's new status
+                metadata: Data associated to the visit
+                snapshot (Optional[sha1_git]): identifier of the snapshot
+                    associated to the visit
+        """
+        origin = self.origin_get({'url': origin}, db=db, cur=cur)
+        if not origin:
+            return
+        origin = origin['id']
+
+        origin_visit = db.origin_visit_get_latest(
+            origin, allowed_statuses=allowed_statuses,
+            require_snapshot=require_snapshot, cur=cur)
+        if origin_visit:
+            return dict(zip(db.origin_visit_get_cols, origin_visit))
+
     @db_transaction(statement_timeout=2000)
     def object_find_by_sha1_git(self, ids, db=None, cur=None):
         """Return the objects found with the given ids.
@@ -1320,10 +1419,13 @@ class Storage():
         """Return origins, either all identified by their ids or all
         identified by tuples (type, url).
 
+        If the url is given and the type is omitted, one of the origins with
+        that url is returned.
+
         Args:
             origin: a list of dictionaries representing the individual
                 origins to find.
-                These dicts have either the keys type and url:
+                These dicts have either the key url (and optionally type):
 
                 - type (FIXME: enum TBD): the origin type ('git', 'wget', ...)
                 - url (bytes): the url the origin points to
@@ -1362,16 +1464,16 @@ class Storage():
             else:
                 raise ValueError(
                     'Either all origins or none at all should have an "id".')
-        elif any(type_ and url for (type_, url) in origin_types_and_urls):
+        elif any(url for (type_, url) in origin_types_and_urls):
             # Lookup per type + URL
-            if all(type_ and url for (type_, url) in origin_types_and_urls):
+            if all(url for (type_, url) in origin_types_and_urls):
                 results = db.origin_get_with(origin_types_and_urls, cur)
             else:
                 raise ValueError(
-                    'Either all origins or none at all should have a '
-                    '"type" and an "url".')
+                    'Either all origins or none at all should have '
+                    'an "url" key.')
         else:  # unsupported lookup
-            raise ValueError('Origin must have either id or (type and url).')
+            raise ValueError('Origin must have either id or url.')
 
         results = [dict(zip(self.origin_keys, result))
                    for result in results]
@@ -1512,6 +1614,12 @@ class Storage():
         """Add an entry for origin origin_id in fetch_history. Returns the id
         of the added fetch_history entry
         """
+        if isinstance(origin_id, str):
+            origin = \
+                self.origin_get([{'url': origin_id}], db=db, cur=cur)
+            if not origin:
+                return
+            origin_id = origin[0]['id']
         fetch_history = {
             'origin': origin_id,
             'date': datetime.datetime.now(tz=datetime.timezone.utc),
